package t1ck.widebase.io.csv.swfx

import java.io. {

  BufferedReader,
  File,
  FileInputStream,
  FilenameFilter,
  FileReader,
  InputStreamReader,
  RandomAccessFile

}

import java.util.zip.GZIPInputStream

import net.liftweb.common. { Loggable, Logger }

import org.joda.time. { LocalDate, LocalDateTime }

import scala.collection.mutable. { ArrayBuffer, Buffer, LinkedHashMap, Map }

import util.control.Breaks. { break, breakable }

import vario.data.Datatype
import vario.file.FileVariantWriter

import widebase.db.column. {

  DoubleColumn,
  IntColumn,
  DateTimeColumn,
  StringColumn,
  TypedColumn

}

import widebase.db.table.Table
import widebase.io.csv.TypeMismatchException
import widebase.io.filter.MagicId

trait TableProcessor extends Logger with Loggable {

  import t1ck.widebase.io.csv.filter.swfx

  import widebase.db

  /** Reads a serie of CSV files sorted and return table.
   *
   * @param delimiter of CSV file
   * @param filter individual processing
   * @param source of CSV files
   * @param side of bid or ask
   * @param from date
   * @param till date
   *
   * @return processed [[widebase.db.Table]]
  */
  def apply(
    delimiter: String,
    filter: Array[String] => Array[String],
    source: String,
    side: String,
    from: LocalDate,
    till: LocalDate): Table = {

    val files = (new File(source)).listFiles(new FilenameFilter {

      def accept(dir: File, name: String) =
        ("""^.*_Daily_""" + side + """_""" + from.toString("yyyy.MM.dd") +
        """_""" + till.toString("yyyy.MM.dd") + """[.](csv|csv.gz)$""").r.
        findFirstIn(name) != None

    } ).toBuffer

    val valueTypes = Array(
      Datatype.DateTime,
      Datatype.Int,
      Datatype.Double,
      Datatype.Double,
      Datatype.Double,
      Datatype.Double,
      Datatype.Double)

    val readers = ArrayBuffer[BufferedReader]()
    val columns = LinkedHashMap[String, TypedColumn[_]]()

    try {

      files.foreach(file =>
        if(file.getName.endsWith(".gz"))
          readers += new BufferedReader(new InputStreamReader(
            new GZIPInputStream(new FileInputStream(file))))
        else
          readers += new BufferedReader(new FileReader(file)))

      var current = -1

      // Ignore records and read label
      current = 0
      var ignoreRecords = 1 // Configurable
      var labelRecord = 1 // Configurable
      var labels: StringColumn = null
      readers.foreach { reader =>

        breakable {

          for(r <- 1 to ignoreRecords) {

            var line = reader.readLine

            if(line == null) {

              warn("Skipped: " + files(current).getName)

              readers.remove(current)
              files.remove(current)
              current -= 1

              break

            }

            if(labels == null && r == labelRecord) {

              labels = new StringColumn

              var header = line.split(delimiter).toBuffer

              header += header.remove(1)
              header.insert(1, "Symbol")
              header.foreach(label => columns += label -> null)

            }
          }
        }

        current += 1

      }

      val records = ArrayBuffer[Buffer[String]]()

      // Read first record and split into fields
      current = 0
      readers.foreach { reader =>

        var line = reader.readLine

        if(line == null || line.isEmpty) {

          warn("Skipped: " + files(current).getName)

          readers.remove(current)
          files.remove(current)
          current -= 1

        } else {

          records += line.split(delimiter).toBuffer
          records.last.insert(1, files(current).getName.takeWhile(c => c != '_'))
          records.last.last.replace(" ", "")

        }

        current += 1

      }

      // Sort and merge records
      var lowestFile = -1
      var sortColumn = 0 //Configurable
      while(!readers.isEmpty) {

        // Sort
        var index = 0
        records.foreach { record =>

          if(
            lowestFile == -1 ||
            records(lowestFile)(sortColumn) > record(sortColumn))
            lowestFile = index

          index += 1

        }

        // Import
        var i = 0
        filter(records(lowestFile).toArray).foreach { value =>

          valueTypes(i) match {

            case Datatype.Double =>
              if(columns.values.toBuffer(i) == null)
                columns(columns.keys.toBuffer(i)) = new DoubleColumn

              columns.values.toBuffer(i).asInstanceOf[DoubleColumn] += value.toDouble

            case Datatype.Int =>
              if(columns.values.toBuffer(i) == null)
                columns(columns.keys.toBuffer(i)) = new IntColumn

              columns.values.toBuffer(i).asInstanceOf[IntColumn] += value.toInt

            case Datatype.DateTime =>
              if(columns.values.toBuffer(i) == null)
                columns(columns.keys.toBuffer(i)) = new DateTimeColumn

              columns.values.toBuffer(i).asInstanceOf[DateTimeColumn] += new LocalDateTime(value.toLong)

          }

          i += 1

        }

        // Debug
//        println(records(lowestFile).reduceLeft(_ + delimiter + _))

        // Read next
        var line = readers(lowestFile).readLine

        if(line == null || line.isEmpty) {

          info("Done: " + files(lowestFile).getName)

          readers.remove(lowestFile)
          files.remove(lowestFile)
          records.remove(lowestFile)
          lowestFile = -1

        } else {

          records(lowestFile) = line.split(delimiter).toBuffer
          records(lowestFile).insert(1, files(lowestFile).getName.takeWhile(c => c != '_'))
          records(lowestFile).last.replace(" ", "")

        }
      }
    } finally {

      readers.foreach(_.close)
      readers.clear

    }

    val table = new Table

    columns foreach { case (name, column) => table ++= (name, column) }

    table

  }

  /** Write a serie of CSV files sorted into database.
   *
   * @param path to database
   * @param table name
   * @param delimiter of CSV file
   * @param filter individual processing
   * @param source of CSV files
   * @param from date
   * @param till date
   * @param segmented path to segment
  */
  def to(
    path: String,
    table: String,
    delimiter: String,
    filter: Array[String] => Array[String],
    source: String,
    from: LocalDate,
    till: LocalDate)
    (implicit segmented: File = null) {

    to(
      new File(path),
      table,
      delimiter,
      filter,
      new File(source),
      from,
      till)(segmented)

  }


  /** Write a serie of CSV files sorted into database.
   *
   * @param path to database
   * @param table name of table
   * @param delimiter of CSV file
   * @param filter individual processing
   * @param source of CSV files
   * @param from date
   * @param till date
   * @param segmented path to segment
  */
  protected def to(
    path: File,
    table: String,
    delimiter: String,
    filter: Array[String] => Array[String],
    source: File,
    from: LocalDate,
    till: LocalDate)
    (implicit segmented: File) {

    val files = source.listFiles(new FilenameFilter {

      def accept(dir: File, name: String) =
        ("""^.*_Ticks_""" + from.toString("yyyy.MM.dd") + """_""" +
        till.toString("yyyy.MM.dd") + """[.](csv|csv.gz)$""").r.
        findFirstIn(name) != None

    } ).toBuffer

    val valueTypes = Array(
      Datatype.DateTime,
      Datatype.Int,
      Datatype.Double,
      Datatype.Double,
      Datatype.Double,
      Datatype.Double)

    val readers = ArrayBuffer[BufferedReader]()
    val writers = ArrayBuffer[FileVariantWriter]()

    def writeRecords(records: Int) {

      writers.foreach { writer =>

        if(writer.isOpen) {

          writer.flush
          writer.mode = Datatype.Byte
          writer.position =
            (MagicId.Column.toString .getBytes(writer.charset)).size + 1

          writer.mode = Datatype.Int
          writer.write(records)

        }
      }
    }

    def releaseWriters {

      writers.foreach(writer =>
        if(writer.isOpen) writer.close)
      writers.clear

    }

    var numberOfRecordsOnPart = 0
    var numberOfRecordsLoaded = 0

    try {

      files.foreach(file =>
        if(file.getName.endsWith(".gz"))
          readers += new BufferedReader(new InputStreamReader(
            new GZIPInputStream(new FileInputStream(file))))
        else
          readers += new BufferedReader(new FileReader(file)))

      var current = -1

      // Ignore records and read label
      current = 0
      var ignoreRecords = 1 // Configurable
      var labelRecord = 1 // Configurable
      var labels: StringColumn = null
      readers.foreach { reader =>

        breakable {

          for(r <- 1 to ignoreRecords) {

            var line = reader.readLine

            if(line == null) {

              warn("Skipped: " + files(current).getName)

              readers.remove(current)
              files.remove(current)
              current -= 1

              break

            }

            if(labels == null && r == labelRecord) {

              labels = new StringColumn

              var header = line.split(delimiter).toBuffer

              header += header.remove(1)
              header.insert(1, "Symbol")
              header.foreach(label => labels += label)

            }
          }
        }

        current += 1

      }

      val records = ArrayBuffer[Buffer[String]]()

      // Read first record and split into fields
      current = 0
      readers.foreach { reader =>

        var line = reader.readLine

        if(line == null || line.isEmpty) {

          warn("Skipped: " + files(current).getName)

          readers.remove(current)
          files.remove(current)
          current -= 1

        } else {

          records += line.split(delimiter).toBuffer
          records.last.insert(1, files(current).getName.takeWhile(c => c != '_'))
          records.last.last.replace(" ", "")

        }

        current += 1

      }

      // Init importer
      class Companion(var writer: FileVariantWriter, var lastEnded: Long)

      var dir: File = null

      def initWriters {

        var valueType = valueTypes.toIterator

        labels.foreach { label =>

          val file = new File(dir.getPath + "/" + label)

          if(file.exists)
            file.delete

          val filename = dir.getPath + "/" + label

          val valueTypeOf = valueType.next

          val channel = new RandomAccessFile(filename, "rw").getChannel
          channel.tryLock
          writers += new FileVariantWriter(channel) {

            override val charset = props.charsets.to

            override def capacity = props.capacities.to
            override def order = props.orders.to

            // Write magic
            mode = Datatype.String
            write(MagicId.Column.toString)

            // Write column type
            mode = Datatype.Byte
            write(valueTypeOf.id.toByte)

            // Write column length
            mode = Datatype.Int
            write(0)

            // Set column value type
            mode = valueTypeOf

          }
        }
      }

      val dbi =
        if(segmented == null)
          db.instance(path.getPath)
        else
          db.instance(segmented.getPath)

      import dbi.tables._

      numberOfRecordsOnPart = 0
      numberOfRecordsLoaded = 0

      var lastInt: Option[Int] = scala.None
      var lastDate: LocalDate = null

      // Sort and merge records
      var lowestFile = -1
      var sortColumn = 0 //Configurable
      while(!readers.isEmpty) {

        // Sort
        var index = 0
        records.foreach { record =>

          if(
            lowestFile == -1 ||
            records(lowestFile)(sortColumn) > record(sortColumn))
            lowestFile = index

          index += 1

        }

        // Import
        val record = filter(records(lowestFile).toArray)

        var partition: String = null

        var newPartition = false

        if(lastDate == null)
          newPartition = true
        else {

          val today = new LocalDate(record.head.toLong)

          if(
            lastDate.getYear != today.getYear ||
            lastDate.getMonthOfYear != today.getMonthOfYear ||
            lastDate.getDayOfMonth != today.getDayOfMonth) {

            info("End of day " + lastDate.toString("yyyy-MM-dd") +
              " passed, +" + numberOfRecordsOnPart + " records")

            newPartition = true

          }
        }

        if(newPartition) {

          lastDate = new LocalDate(record.head.toLong)
          partition = lastDate.toString("yyyy-MM-dd")

        }

        if(partition != null) {

          writeRecords(numberOfRecordsOnPart)
          releaseWriters

          // Set column labels
          save.col(table, ".d", labels)(partition)

          dir =
            if(segmented == null)
              new File(path.getPath + "/" + partition + "/" + table)
            else
              new File(segmented.getPath + "/" + partition + "/" + table)

          initWriters
          numberOfRecordsOnPart = 0

        }

        var i = 0
        record.foreach { value =>

          valueTypes(i) match {

            case Datatype.Double => writers(i).write(value.toDouble)
            case Datatype.Int => writers(i).write(value.toInt)
            case Datatype.DateTime => writers(i).write(new LocalDateTime(value.toLong))

          }

          i += 1

        }

        numberOfRecordsOnPart += 1
        numberOfRecordsLoaded += 1

        // Debug
//        record(0) = new LocalDateTime(record.head.toLong).toString
//        println(record.reduceLeft(_ + delimiter + _))

        // Read next
        var line = readers(lowestFile).readLine

        if(line == null || line.isEmpty) {

          info("Done: " + files(lowestFile).getName)

          readers.remove(lowestFile)
          files.remove(lowestFile)
          records.remove(lowestFile)
          lowestFile = -1

        } else {

          records(lowestFile) = line.split(delimiter).toBuffer
          records(lowestFile).insert(1, files(lowestFile).getName.takeWhile(c => c != '_'))
          records(lowestFile).last.replace(" ", "")

        }
      }
    } finally {

      readers.foreach(_.close)
      readers.clear

      writeRecords(numberOfRecordsOnPart)
      releaseWriters

      info("All data written succesfully, number of records written: " +
        numberOfRecordsLoaded)

    }
  }
}

